@@ -1,5 +1,5 @@
 import asyncio
-import collections
+import collections.abc
 import functools
 import inspect
 import time
@@ -82,7 +82,7 @@ async def test_wrapper(item):
     async def do_teardowns():
         item.start_teardown = time.time()
         for teardown in teardowns:
-            if isinstance(teardown, collections.Iterator):
+            if isinstance(teardown, collections.abc.Iterator):
                 try:
                     teardown.__next__()
                 except StopIteration:
@@ -194,11 +194,14 @@ def pytest_runtestloop(session):
     item_by_coro = {}
     tasks = []
     for item in session.items:
-        markers = [m.name for m in item.own_markers]
+        markers = {m.name: m for m in item.own_markers}
 
         if "skip" in markers:
             continue
 
+        if "skipif" in markers and markers["skipif"].args[0]:
+            continue
+
         # Coerce into a task
         if "asyncio_cooperative" in markers:
             try:
